/*
 * Copyright (C) 2005 Alex Murray <murray.alex@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include "sensors-applet-gconf-keys.h"
#include "sensors-list-i2c-proc.h"
#include "sensors-list.h"

#define I2C_PROC_BASE_DIR "/proc/sys/dev/sensors"
#define ID "i2c-proc"

/* for error handling */
#define I2C_PROC_DEVICE_FILE_ERROR (sensors_list_i2c_proc_device_file_error_quark())

enum {
	I2C_PROC_DEVICE_FILE_OPEN_ERROR,
	I2C_PROC_DEVICE_FILE_READ_ERROR
};

static SensorsList *sensors_list_i2c_proc_sensors_list_new(SensorsApplet *sensors_applet) {
	SensorsList *proc_sensors_list;
	proc_sensors_list = g_new0(SensorsList, 1);
	proc_sensors_list->sensors_applet = sensors_applet;
	proc_sensors_list->id = ID;
	/* set update and error functions */
	proc_sensors_list->update_labels = sensors_list_i2c_proc_update_labels;
	sensors_list_init(proc_sensors_list);

	sensors_applet->sensors_lists = g_slist_append(sensors_applet->sensors_lists, proc_sensors_list);
	return proc_sensors_list;
}

static void sensors_list_i2c_proc_setup_from_gconf(SensorsApplet *sensors_applet) {
	SensorsList *proc_sensors_list = sensors_list_i2c_proc_sensors_list_new(sensors_applet);
	sensors_list_setup_from_gconf(proc_sensors_list);
}
	
static void sensors_list_i2c_proc_add_sensor(SensorsApplet *sensors_applet, const gchar *path) {
	SensorsList *sensors_list;
	gchar *filename;
	gchar *label;
	gboolean enable;
	guint sensor_type;
	GtkTreeIter iter;
	
	sensors_list = sensors_applet_get_sensors_list(sensors_applet, ID);
	if (sensors_list == NULL) {
		sensors_list = sensors_list_i2c_proc_sensors_list_new(sensors_applet);
	}

	filename = g_path_get_basename(path);
	label = g_strdup(filename);

	/* setup temp2 as CPU sensor and enable it */
	if (g_ascii_strcasecmp(filename, "temp2") == 0) {
		sensor_type = I2C_PROC_TEMP_SENSOR;
		label = g_strdup(_("CPU"));
		enable = TRUE;
	} else {
		switch(filename[0]) {
		case 'c':     /* currents are "curr?" */
			sensor_type = I2C_PROC_CURRENT_SENSOR;
			break;
		case 'f':     /* fans are "fan?" */
			sensor_type = I2C_PROC_FAN_SENSOR;
			break;
		case 'i':     /* voltages are "in?" */
			sensor_type = I2C_PROC_VOLTAGE_SENSOR;
			break;
		case 't':     /* temps are "temp?" */
			sensor_type = I2C_PROC_TEMP_SENSOR;
			break;
		case 'v':     /* vids are just vid */
			sensor_type = I2C_PROC_VID_SENSOR;
			break;
		default:
			/* SHOULDN'T BE ABLE
			 * TO GET HERE!! */
			g_assert_not_reached();
		}
		/* disable all other sensors */
		enable = FALSE;
	}
	gtk_list_store_append(sensors_list->list, &iter);
	gtk_list_store_set(sensors_list->list,
			   &iter,
			   PATH_COLUMN, path,
			   FILENAME_COLUMN, filename,
			   LABEL_COLUMN, label,
			   ENABLE_COLUMN, enable,
			   ALARM_VALUE_COLUMN, 0.0,
			   ALARM_TYPE_COLUMN, ALARM_WHEN_VALUE_GREATER_THAN_THRESHOLD,
			   ALARM_ENABLE_COLUMN, FALSE,
			   ALARM_COMMAND_COLUMN, "",
			   ALARM_TIMEOUT_COLUMN, 0,
			   ALARM_TIMEOUT_INDEX_COLUMN, -1,
			   SENSOR_TYPE_COLUMN, sensor_type,
			   MULTIPLIER_COLUMN, 1.0,
			   OFFSET_COLUMN, 0.0,
			   -1);

	g_free(filename);
	g_free(label);
}	

/* recursive function to find sensors in a given path - once found
   will call function to add sensor to the sensors_list */
static void sensors_list_i2c_proc_find_sensors(SensorsApplet *sensors_applet, const gchar *path) {
	GDir *dir;
	FILE *fp;
	gchar *filename;
	const gchar* new_file;
	gchar *new_path;
	if (g_file_test(path, G_FILE_TEST_IS_REGULAR)) {
		filename = g_path_get_basename(path);
		/* see if filename starts with any of the sensor
		   prefixes */
		if (g_strrstr(filename, "curr") == filename || (g_strrstr(filename, "fan") == filename && g_strrstr(filename, "fan_div") == NULL) || g_strrstr(filename, "in") == filename || g_strrstr(filename, "temp") == filename || g_strrstr(filename, "vid") == filename) {
			/* also test can actually open file for
			   reading */
			if (fp = fopen(path, "r")) {
				fclose(fp);
				/* now add to sensor list */
				sensors_list_i2c_proc_add_sensor(sensors_applet, path);
			}
		}
		g_free(filename);
		
	}
	/* if is a directory (but not a symlinked dir as this
	   will lead us in circular loops) descend into it and look
	   for a sensor dir
	*/
	if (g_file_test(path, G_FILE_TEST_IS_DIR) && !g_file_test(path, G_FILE_TEST_IS_SYMLINK)) {
		dir = g_dir_open(path, 0, NULL);
		if (dir != NULL) {
			while(new_file = g_dir_read_name(dir)) {
				new_path = g_strdup_printf("%s/%s", path, new_file);
				sensors_list_i2c_proc_find_sensors(sensors_applet, new_path);
				g_free(new_path);
			}
			g_dir_close(dir);
		}
	}
}

/* to be called to setup for proc sensors */
void sensors_list_i2c_proc_init(SensorsApplet *sensors_applet) {
	/* if not setup find sensors manually */
	if (!panel_applet_gconf_get_bool(sensors_applet->applet, ID "/" IS_SETUP, NULL)) {
		/* call function to recursively look for sensors
		   starting at the defined base directory */
		sensors_list_i2c_proc_find_sensors(sensors_applet, I2C_PROC_BASE_DIR);
	} else {
		if (sensors_list_gconf_paths_valid(sensors_applet, ID)) {
			sensors_list_i2c_proc_setup_from_gconf(sensors_applet);
		} else {
			/* alert the user and restart from scratch */
			panel_applet_gconf_set_bool(sensors_applet->applet, ID "/" IS_SETUP, FALSE, NULL);
			sensors_list_gconf_error_occurred(ID);
			sensors_list_i2c_proc_find_sensors(sensors_applet, I2C_PROC_BASE_DIR);
		}
	}
}

static GQuark sensors_list_i2c_proc_device_file_error_quark(void) {
	static GQuark quark = 0;

	if (quark == 0) {
		quark = g_quark_from_static_string(ID "-device-file-error");
	}

	return quark;
}


static gfloat sensors_list_i2c_proc_get_sensor_value(SensorsList *sensors_list,
						     GtkTreeIter *iter,
						     GError **error) {
	gchar *sensor_path = NULL;
	gdouble sensor_multiplier, sensor_offset;
	I2cProcSensorType sensor_type;

	/* to open and access the value of each sensor */
	FILE *fp;
	gfloat float1, float2, float3;
	gint int1, int2, int3;

	gfloat sensor_value;

	gtk_tree_model_get(GTK_TREE_MODEL(sensors_list->list), 
			   iter,
			   PATH_COLUMN, &sensor_path,
			   SENSOR_TYPE_COLUMN, &sensor_type,
			   MULTIPLIER_COLUMN, &sensor_multiplier,
			   OFFSET_COLUMN, &sensor_offset,
			   -1);

	if (NULL == (fp = fopen(sensor_path, "r"))) {
		g_set_error(error, I2C_PROC_DEVICE_FILE_ERROR, I2C_PROC_DEVICE_FILE_OPEN_ERROR, "Error opening sensor device file %s", sensor_path);
		g_free(sensor_path);
		return;
	}

	switch (sensor_type) {
	case I2C_PROC_CURRENT_SENSOR:

		if (fscanf(fp, "%f %f %f", &float1, &float2, &float3) != 3) {
			g_set_error(error, I2C_PROC_DEVICE_FILE_ERROR, I2C_PROC_DEVICE_FILE_READ_ERROR, "Error reading from sensor device file %s", sensor_path);
			g_free(sensor_path);
			fclose(fp);
			return;
		}
		
		sensor_value = float3;
		break;
		
	case I2C_PROC_FAN_SENSOR:
		if (fscanf(fp, "%d %d", &int1, &int2) != 2) {
			g_set_error(error, I2C_PROC_DEVICE_FILE_ERROR, I2C_PROC_DEVICE_FILE_READ_ERROR, "Error reading from sensor device file %s", sensor_path);
			g_free(sensor_path);
			fclose(fp);
			return;
		}

		sensor_value = (gfloat)int2;
		break;
		
	case I2C_PROC_VOLTAGE_SENSOR:
		if (fscanf(fp, "%f %f %f", &float1, &float2, &float3) != 3) {
			g_set_error(error, I2C_PROC_DEVICE_FILE_ERROR, I2C_PROC_DEVICE_FILE_READ_ERROR, "Error reading from sensor device file %s", sensor_path);
			g_free(sensor_path);
			fclose(fp);
			return;
		}
		sensor_value = float3;
		break;
		
	case I2C_PROC_TEMP_SENSOR:
		if (fscanf(fp, "%f %f %f", &float1, &float2, &float3) != 3) {
			g_set_error(error, I2C_PROC_DEVICE_FILE_ERROR, I2C_PROC_DEVICE_FILE_READ_ERROR, "Error reading from sensor device file %s", sensor_path);
			g_free(sensor_path);
			fclose(fp);
			return;
		}
		sensor_value = float3;		
		if (panel_applet_gconf_get_bool(sensors_list->sensors_applet->applet, FAHRENHEIT, NULL)) {
			/* Convert to fahrenheit */
			sensor_value = 32.0 + 9.0 * sensor_value / 5.0;
		}
		break;

	case I2C_PROC_VID_SENSOR:
		/* want first value in file as float */
		if (fscanf(fp, "%f", &float1) != 1) {
			g_set_error(error, I2C_PROC_DEVICE_FILE_ERROR, I2C_PROC_DEVICE_FILE_READ_ERROR, "Error reading from sensor device file %s", sensor_path);
			g_free(sensor_path);
			fclose(fp);
			return -1.0;
		}
		sensor_value = float1;
		break;
	default:
		g_assert_not_reached();
	} /* end switch */
	g_free(sensor_path);
	fclose(fp);

	return ((sensor_value * sensor_multiplier) + sensor_offset);
}

void sensors_list_i2c_proc_update_labels(SensorsList *sensors_list) {
	/* ensure we have a data structure to work from */
	g_assert(sensors_list != NULL);
	g_assert(sensors_list->list != NULL);

	/* to iterate through the sensors data structure
	   (sensors_list) */
	GtkTreeIter iter;
	gboolean not_end_of_list = TRUE;

	gchar *sensor_path = NULL;
	gchar *sensor_label = NULL;
	gboolean sensor_enabled;
	gdouble sensor_alarm_value;
	gboolean sensor_alarm_enabled;
	AlarmType sensor_alarm_type;
	I2cProcSensorType sensor_type;

	gfloat sensor_value;

	/* to build the list of labels as we go */
	gchar *label_text = NULL;
	gchar *value_text = NULL;
	gchar *font_size = NULL;

	GSList *current_label = NULL;
	GSList *current_value = NULL;

	/* get the first label in the list */
	current_label = sensors_list->labels;
	current_value = sensors_list->values;
	
	/* now step through the GtkListStore sensors_list to
	   find which sensors are enabled */
	for (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(sensors_list->list), &iter); not_end_of_list; not_end_of_list = gtk_tree_model_iter_next(GTK_TREE_MODEL(sensors_list->list), &iter)) {
		gtk_tree_model_get(GTK_TREE_MODEL(sensors_list->list), 
				   &iter,
				   PATH_COLUMN, &sensor_path,
				   LABEL_COLUMN, &sensor_label,
				   ENABLE_COLUMN, &sensor_enabled,
				   ALARM_VALUE_COLUMN, &sensor_alarm_value,
				   ALARM_TYPE_COLUMN, &sensor_alarm_type,
				   ALARM_ENABLE_COLUMN, &sensor_alarm_enabled,
				   SENSOR_TYPE_COLUMN, &sensor_type,
				   -1);

		if (sensor_enabled) {
			/* get font size */
			font_size = sensors_list_get_font_size(sensors_list);

			GError *error = NULL;
			sensor_value = sensors_list_i2c_proc_get_sensor_value(sensors_list, &iter, &error);
			if (error != NULL) {
				value_text = g_markup_printf_escaped("<span size=\"%s\">%s</span>", font_size, error->message);
				g_error_free(error);
			} else { 
				switch (sensor_type) {
				case I2C_PROC_CURRENT_SENSOR:
					value_text = g_strdup_printf("%4.2f%s", sensor_value, (panel_applet_gconf_get_bool(sensors_list->sensors_applet->applet, SHOW_UNITS, NULL) ? _("A") : ""));
					break;
				case I2C_PROC_FAN_SENSOR:
					value_text = g_strdup_printf("%4.0f%s", sensor_value, (panel_applet_gconf_get_bool(sensors_list->sensors_applet->applet, SHOW_UNITS, NULL) ? _("RPM") : ""));
					
					break;
				case I2C_PROC_VOLTAGE_SENSOR:
					value_text = g_strdup_printf("%4.2f%s", sensor_value, (panel_applet_gconf_get_bool(sensors_list->sensors_applet->applet, SHOW_UNITS, NULL) ? _("V") : ""));
					
					break;
				case I2C_PROC_TEMP_SENSOR:
					if (panel_applet_gconf_get_bool(sensors_list->sensors_applet->applet, FAHRENHEIT, NULL)) {
						value_text = g_strdup_printf("%3.1f%s", sensor_value, (panel_applet_gconf_get_bool(sensors_list->sensors_applet->applet, SHOW_UNITS, NULL) ? "\302\260F" : ""));
					} else {
						value_text = g_strdup_printf("%3.1f%s", sensor_value, (panel_applet_gconf_get_bool(sensors_list->sensors_applet->applet, SHOW_UNITS, NULL) ? "\302\260C" : ""));
					}
					
					break;
				case I2C_PROC_VID_SENSOR:
					value_text = g_strdup_printf("%4.2f%s", sensor_value, (panel_applet_gconf_get_bool(sensors_list->sensors_applet->applet, SHOW_UNITS, NULL) ? _("V") : ""));
					break;
				default:
					g_assert_not_reached();
				} /* end switch */
				
				gchar *old_value_text = value_text;
				
		
				if (sensor_alarm_enabled) {
					switch (sensor_alarm_type) {
					case ALARM_WHEN_VALUE_GREATER_THAN_THRESHOLD:
						if (sensor_value >= sensor_alarm_value) {
							/* make value
							 * text red with
							 correct font
							 size and do
							 alarm command */
							value_text = g_markup_printf_escaped("<span foreground=\"#FF0000\" size=\"%s\">%s</span>", font_size, old_value_text);
							sensors_list_alarm_on(sensors_list, &iter);
						} else {
							/* just do font size */
							value_text = g_markup_printf_escaped("<span size=\"%s\">%s</span>", font_size, old_value_text);
							sensors_list_alarm_off(sensors_list, &iter);
						}				
						break;
					case ALARM_WHEN_VALUE_LESS_THAN_THRESHOLD:
						if (sensor_value <= sensor_alarm_value) {
							/* make value
							 * text red with
							 correct font
							 size and do
							 alarm command */
							value_text = g_markup_printf_escaped("<span foreground=\"#FF0000\" size=\"%s\">%s</span>", font_size, old_value_text);
							sensors_list_alarm_on(sensors_list, &iter);
						} else {
							/* just do font size */
							value_text = g_markup_printf_escaped("<span size=\"%s\">%s</span>", font_size, old_value_text);
							sensors_list_alarm_off(sensors_list, &iter);
						}				
						break;
					default: 
						g_assert_not_reached();
					} // end switch(sensor_alarm_type)
				} else {
					/* just do font size */
					value_text = g_markup_printf_escaped("<span size=\"%s\">%s</span>", font_size, old_value_text);
					sensors_list_alarm_off(sensors_list, &iter);
				}
				g_free(old_value_text);
			} /* end of else from if (error) */

			sensors_list_add_label_to_list(&(sensors_list->values), &current_value, value_text);
			g_free(value_text);
			
			/* advance to next element in list */
			current_value = g_slist_next(current_value);

			/* create label text with font size
			 * formatting */
			label_text = g_markup_printf_escaped("<span size=\"%s\">%s</span>", font_size, sensor_label);
			sensors_list_add_label_to_list(&(sensors_list->labels), &current_label, label_text);
			g_free(label_text);
			current_label = g_slist_next(current_label);
			
			g_free(font_size);
		} // end if (sensor_enabled)
		g_free(sensor_path);
		g_free(sensor_label);
				
	}
	/* now remove any remaining elements in either list to free up
	   memory */
	sensors_list_crop_list(&(sensors_list->values), &current_value);
	sensors_list_crop_list(&(sensors_list->labels), &current_label);
	
	
}
